/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * This file implements the Port Role Selection STM of 17.22.
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp.h"
#include "br_rstp_stm.h"

static void step(struct net_bridge*, int*);
br_rstp_stm_bridge_function br_rstp_stm_port_role_selection = &step;

static void clearReselectBridge(struct net_bridge*);
static void setSelectedBridge(struct net_bridge*);
static void updtRoleDisabledBridge(struct net_bridge*);
static void updtRolesBridge(struct net_bridge*);

typedef enum {
	BEGIN = BR_RSTP_STM_STATE_BEGIN,
	INIT_BRIDGE,
	ROLE_SELECTION
} states;

static
void
enter(struct net_bridge *br, states *state, states target)
{
	switch (target) {
		case INIT_BRIDGE:
			updtRoleDisabledBridge(br);
			break;

		case ROLE_SELECTION:
			clearReselectBridge(br);
			updtRolesBridge(br);
			setSelectedBridge(br);
			break;

		default:
			printk(KERN_INFO "%s: bad transition target in %s: %d -> %d\n",
			       br->dev.name, __FILE__, *state, target);
			return;
	}
	*state = target;
}

static
void
step(struct net_bridge *br, int *state_val)
{
	states state = *(states*) state_val;
#define enter(target) enter(br, (states*) state_val, target)
	switch (state) {
		case BEGIN:
			enter(INIT_BRIDGE);
			break;

		case INIT_BRIDGE:
			enter(ROLE_SELECTION);
			break;

		case ROLE_SELECTION:
		{
			struct net_bridge_port *p = br->port_list;
			while (p != NULL) {
				if (p->reselect) {
					enter(ROLE_SELECTION);
					break;
				}
				p = p->next;
			}
			break;
		}

		default:
			printk(KERN_INFO "%s: bad state in %s: %d; resetting to BEGIN\n",
			       br->dev.name, __FILE__, state);
			enter(BEGIN);
	}
#undef enter
}

/* 17.19.1 */
static
void
clearReselectBridge(struct net_bridge *br)
{
	struct net_bridge_port *p = br->port_list;
	while (p != NULL) {
		p->reselect = FALSE;
		p = p->next;
	}
}

/* 17.19.12 */
static
void
setSelectedBridge(struct net_bridge *br)
{
	struct net_bridge_port *p;

	/* abort without doing anything if any port has reselect set */
	p = br->port_list;
	while (p != NULL) {
		if (p->reselect)
			return;
		p = p->next;
	}

	/* no port has reselect set: set selected on all ports */
	p = br->port_list;
	while (p != NULL) {
		p->selected = TRUE;
		p = p->next;
	}
}

/* 17.19.20 */
static
void
updtRoleDisabledBridge(struct net_bridge *br)
{
	struct net_bridge_port *p = br->port_list;
	while (p != NULL) {
		p->selectedRole = BR_ROLE_DISABLED;
		p = p->next;
	}
}

/* 17.19.21 */
static
void
updtRolesBridge(struct net_bridge *br)
{
	struct net_bridge_port *p;
	port_id prev_root = br->rootPortId;

	/* a) through d) */
	/* determine best root path priority vector and associated times */
	{
		struct net_bridge_priority_vector best_rppv = br->BridgePriority;
		struct net_bridge_times best_times = br->BridgeTimes;

		p = br->port_list;
		while (p != NULL) {
			if (p->portEnabled &&
			    p->infoIs == BR_INFO_RECEIVED)
			{  /* ADDED for carrier detection */
				struct net_bridge_priority_vector rppv;

				/* a) 17.4.2.2, root path priority vector */
				rppv = p->msgPriority;
				rppv.RootPathCost += p->PortPathCost;

				/* c) update best_{rppv,times} if new best vector */
				if (memcmp(&rppv.DesignatedBridgeID, &br->BridgePriority.DesignatedBridgeID, 8) &&
				    br_stp_compare_priority_vectors(&rppv, &best_rppv, TRUE) < 0)
				{
					unsigned long l;

					best_rppv = rppv;

					/* b) + d) (2) */
					best_times = p->portTimes;
					l = best_times.MaxAge / 16 > 1*HZ ? best_times.MaxAge / 16 : 1*HZ;
					// round to nearest whole second
					best_times.MessageAge = ((best_times.MessageAge + l + HZ/2) / HZ) * HZ;
				}
			}

			p = p->next;
		}
		/* c) (continued) */
		br->rootPriority.RootBridgeID = best_rppv.RootBridgeID;
		br->rootPriority.RootPathCost = best_rppv.RootPathCost;
		br->rootPriority.DesignatedBridgeID = best_rppv.DesignatedBridgeID;
		br->rootPriority.DesignatedPortID = best_rppv.DesignatedPortID;
		if (br->rootPortId != best_rppv.BridgePortID) {
			if (best_rppv.BridgePortID == 0)
				printk(KERN_INFO "%s: assuming ROOT BRIDGE status\n", br->dev.name);
			else if (br->rootPortId == 0) {
				printk(KERN_INFO "%s: losing ROOT BRIDGE status to ", br->dev.name);
				br_dump_bridge_id(&best_rppv.RootBridgeID);
				printk("\n");
			}
		}
		br->rootPortId = best_rppv.BridgePortID;

		/* d) (continued) */
		br->rootTimes = best_times;
	}
//printk("RPV: ");br_dump_priority_vector(&br->rootPriority);printk(", root=%04x\n", br->rootPortId);

	/* e) through l) */
	p = br->port_list;
	while (p != NULL) {
		/* e) + 17.4.2.2, designated priority vector */
		if (memcmp(&br->BridgeIdentifier, &br->rootPriority.RootBridgeID, 8) < 0) {
			p->designatedPriority.RootBridgeID = br->BridgeIdentifier;
			p->designatedPriority.RootPathCost = 0;
		} else {
			p->designatedPriority.RootBridgeID = br->rootPriority.RootBridgeID;
			p->designatedPriority.RootPathCost = br->rootPriority.RootPathCost;
		}
		p->designatedPriority.DesignatedBridgeID = br->BridgeIdentifier;
		p->designatedPriority.DesignatedPortID = p->portId;
//printk("DPV: ");br_dump_priority_vector(&p->designatedPriority);printk(", infoIs=%d\n", p->infoIs);

		/* f) */
		p->designatedTimes = br->rootTimes;

		/* g) through l) */
		if (p->infoIs == BR_INFO_DISABLED)
			p->selectedRole = BR_ROLE_DISABLED;  /* g) */
		else if (p->infoIs == BR_INFO_AGED) {
			/* h) */
			p->updtInfo = TRUE;
			p->selectedRole = BR_ROLE_DESIGNATED;
		} else if (p->infoIs == BR_INFO_MINE) {
			/* i) */
			p->selectedRole = BR_ROLE_DESIGNATED;
			if (br_stp_compare_priority_vectors(&p->portPriority, &p->designatedPriority, TRUE))
				p->updtInfo = TRUE;
			else {
				struct net_bridge_port *q = br->port_list;
				while (q != NULL) {
					if (br->rootPortId == q->portId) {
						if (br_stp_compare_times(&p->portTimes, &q->portTimes))
							p->updtInfo = TRUE;
						break;
					}
					q = q->next;
				}
			}
		} else if (p->infoIs == BR_INFO_RECEIVED) {
			/* j) */
			if (br->rootPortId == p->portId) {
				p->selectedRole = BR_ROLE_ROOT;
				p->updtInfo = FALSE;

#if 0  /* NON-FUNCTIONAL */
				/*
				 * The following code implements the optimization suggestion of
				 * 17.10.  This optimization consists in switching the current
				 * root port's FDB entries to the new root if the latter was an
				 * alternate port that can immediately be made forwarding.  To
				 * check the latter, the relevant part of the STM transitions'
				 * conditions of 17.23 from ROOT_PORT to ROOT_LEARN and then
				 * ROOT_FORWARD are evaluated.
				 */
				if (p->portId != prev_root &&
				    p->role == BR_ROLE_ALTERNATE &&
				    br->ForceVersion >= 2 &&
				    br_timer_has_expired(&p->rbWhile))
				{
					struct net_bridge_port *q;
					int reRooted = TRUE;

					/*
					 * Calculate reRooted as used in the Root Port Role
					 * Transitions STM of 17.23.
					 */
					q = br->port_list;
					while (q != NULL) {
						if (q != p && q->portId != prev_root &&
						    !br_timer_has_expired(&q->rrWhile))
						{
							reRooted = FALSE;
							break;
						}
						q = q->next;
					}

					if (reRooted) {
						struct net_bridge_port *prev = NULL;

						q = br->port_list;
						while (q != NULL) {
							if (q->portId == prev_root) {
								prev = q;
								break;
							}
							q = q->next;
						}

						if (prev != NULL) {
							printk(KERN_INFO "%s: rapid root switch: transferring FDB entries from port %d to port %d\n",
							       p->br->dev.name, prev->port_no, p->port_no);
							br_fdb_switch_ports(br, prev, p);
						}
					}
				}
#endif
			} else {  /* br->rootPortId != p->portId */
				/* k) and l) */
#if 1  /* XXX: not in the spec but needed to work--where's the bug? */
				if (memcmp(&p->portPriority.RootBridgeID, &br->rootPriority.RootBridgeID, 8) > 0) {
					p->selectedRole = BR_ROLE_DESIGNATED;
					p->updtInfo = TRUE;
				} else
#else
				if (br_stp_compare_priority_vectors(&p->designatedPriority, &p->portPriority, FALSE) <= 0)
#endif
				{
					int local = !memcmp(&p->portPriority.DesignatedBridgeID, &br->BridgeIdentifier, 8);

					if (local) {
						struct net_bridge_port *q;

						local = FALSE;
						q = br->port_list;
						while (q != NULL && !local) {
							if (p->portPriority.DesignatedPortID == q->portId)
								local = TRUE;
							q = q->next;
						}
					}

					p->selectedRole = !local ? BR_ROLE_ALTERNATE  /* k) */
					                         : BR_ROLE_BACKUP;    /* l) */
					p->updtInfo = FALSE;
#if 0  /* XXX: not in the spec but needed to work--where's the bug? */
				} else {
					/*
					 * NOTE: Without this, no assignment would be made, which
					 *       may result in, e.g., duplicate root ports.
					 */
					if (memcmp(&p->portPriority.RootBridgeID, &br->rootPriority.RootBridgeID, 8) > 0) {
						p->selectedRole = BR_ROLE_DESIGNATED;
						p->updtInfo = TRUE;
					} else {
						p->selectedRole = BR_ROLE_ALTERNATE;
						p->updtInfo = FALSE;
					}
#endif
				}
			}
		}

		p = p->next;
	}
}
