/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * This file implements the Port Transmit STM of 17.27.
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp.h"
#include "br_rstp_stm.h"

static void step(struct net_bridge_port*, int*);
br_rstp_stm_port_function br_rstp_stm_port_transmit = &step;

static void txConfig(struct net_bridge_port*);
static void txRstp(struct net_bridge_port*);
static void txTcn(struct net_bridge_port*);

typedef enum {
	BEGIN = BR_RSTP_STM_STATE_BEGIN,
	TRANSMIT_INIT,
	IDLE,
	TRANSMIT_PERIODIC,
	TRANSMIT_CONFIG,
	TRANSMIT_TCN,
	TRANSMIT_RSTP
} states;

static
void
enter(struct net_bridge_port *p, states *state, states target)
{
	switch (target) {
		case TRANSMIT_INIT:
			p->newInfo = FALSE;
			br_timer_clear(&p->helloWhen);
			p->txCount = 0;
			break;

		case IDLE:
			// nothing to be done
			break;

		case TRANSMIT_PERIODIC:
			p->newInfo =
				p->newInfo ||
				p->role == BR_ROLE_DESIGNATED ||
				(p->role == BR_ROLE_ROOT && br_timer_is_running(&p->tcWhile));
			br_timer_set(&p->helloWhen, p->portTimes.HelloTime);
			break;

		case TRANSMIT_CONFIG:
			p->newInfo = FALSE;
			txConfig(p);
			p->txCount++;
			p->tcAck = FALSE;
			break;

		case TRANSMIT_TCN:
			p->newInfo = FALSE;
			txTcn(p);
			p->txCount++;
			break;

		case TRANSMIT_RSTP:
			p->newInfo = FALSE;
			txRstp(p);
			p->txCount++;
			p->tcAck = FALSE;
			break;

		default:
			printk(KERN_INFO "%s: bad transition target in %s: %d -> %d\n",
			       p->br->dev.name, __FILE__, *state, target);
			return;
	}
	*state = target;
}

static
void
step(struct net_bridge_port *p, int *state_val)
{
	states state = *(states*) state_val;
#define enter(target) enter(p, (states*) state_val, target)
	switch (state) {
		case BEGIN:
			enter(TRANSMIT_INIT);
			break;

		case TRANSMIT_INIT:
		case TRANSMIT_PERIODIC:
		case TRANSMIT_CONFIG:
		case TRANSMIT_TCN:
		case TRANSMIT_RSTP:
			enter(IDLE);
			break;

		case IDLE:
			if (p->selected && !p->updtInfo) {
				if (br_timer_has_expired(&p->helloWhen))
					enter(TRANSMIT_PERIODIC);
				else if (p->newInfo && p->txCount < p->br->TxHoldCount) {
					if (p->sendRSTP) {
						if (p->role == BR_ROLE_ROOT || p->role == BR_ROLE_DESIGNATED)
							enter(TRANSMIT_RSTP);
					} else if (!p->sendRSTP) {
						if (p->role == BR_ROLE_ROOT)
							enter(TRANSMIT_TCN);
						else if (p->role == BR_ROLE_DESIGNATED)
							enter(TRANSMIT_CONFIG);
					}
				}
			}
			break;

		default:
			printk(KERN_INFO "%s: bad state in %s: %d; resetting to BEGIN\n",
			       p->br->dev.name, __FILE__, state);
			enter(BEGIN);
	}
#undef enter
}

/* 17.19.15 */
static
void
txConfig(struct net_bridge_port *p)
{
	struct net_bridge_bpdu bpdu;

	memset(&bpdu, 0, sizeof(bpdu));
	bpdu.type = BR_BPDU_TYPE_CFG;

	bpdu.priority = p->portPriority;
	bpdu.times    = p->portTimes;

	if (br_timer_is_running(&p->tcWhile))
		bpdu.tc = TRUE;
	bpdu.tcAck = p->tcAck;

	br_stp_send_bpdu(p, &bpdu);
}

/* 17.19.16 */
static
void
txRstp(struct net_bridge_port *p)
{
	struct net_bridge_bpdu bpdu;

	memset(&bpdu, 0, sizeof(bpdu));
	bpdu.type = BR_BPDU_TYPE_RST;

	bpdu.priority = p->portPriority;
	bpdu.times    = p->portTimes;

	if (br_timer_is_running(&p->tcWhile))
		bpdu.tc = TRUE;
	bpdu.tcAck = FALSE;

	switch (p->role) {
		case BR_ROLE_ALTERNATE:
		case BR_ROLE_BACKUP:
			bpdu.role = BR_BPDU_ROLE_ALT_BKP;
			break;

		case BR_ROLE_ROOT:
			bpdu.role = BR_BPDU_ROLE_ROOT;
			break;

		case BR_ROLE_DESIGNATED:
			bpdu.role = BR_BPDU_ROLE_DESIGNATED;
			break;

		default:
			bpdu.role = BR_BPDU_ROLE_UNKNOWN;
	}

	bpdu.proposal = p->proposing;
	bpdu.learning = p->learning;
	bpdu.forwarding = p->forwarding;
	bpdu.agreement = p->synced;

	br_stp_send_bpdu(p, &bpdu);
}

/* 17.19.17 */
static
void
txTcn(struct net_bridge_port *p)
{
	struct net_bridge_bpdu bpdu;

	memset(&bpdu, 0, sizeof(bpdu));
	bpdu.type = BR_BPDU_TYPE_TCN;

	br_stp_send_bpdu(p, &bpdu);
}
